@@ -3,7 +3,7 @@ import json
 import operator
 import os
 import struct
-from typing import Dict, Iterable, List, Optional, Tuple
+from typing import Dict, IO, Iterable, List, Optional, Tuple
 
 import attr
 import marshmallow
@@ -33,7 +33,7 @@ class SafeTensors:
     ----------
     file: str, file-like, or Path
         File to operate on.
-    mode: {"r", "r+", "w+", "c"}
+    mode: {"r", "r+", "x+", "w+", "c"}
         r: Open existing file for reading only.
         r+: Open existing file for reading and writing.
         w+: Create or overwrite existing file for reading and writing.
@@ -45,19 +45,33 @@ class SafeTensors:
         to do so.
     """
 
-    _valid_modes = frozenset(("r", "r+", "w+", "c"))
+    _valid_modes = frozenset(("r", "r+", "x+", "w+", "c"))
 
-    file: object = attr.ib()
+    file: IO = attr.ib()
     mode: str = attr.ib()
     sparse: bool = attr.ib(default=None)
-    settings: "FileMapSettings" = attr.ib(
-        factory=lambda: FileMapSettings(), kw_only=True
-    )
+    settings: "Settings" = attr.ib(factory=lambda: Settings(), kw_only=True)
 
     map: "FileMap"
 
     tensors: "Dict[str, SafeTensorInfo]"
-    _close_file = False
+    _close_on_exit: bool = False
+
+    def _open_file(self):
+        self.file = open(self.file, self.mode.replace("c", "r") + "b")
+
+    def __attrs_post_init__(self):
+        if not hasattr(self.file, "read"):
+            # the user passed in a Path or str, so we need to make sure to close it
+            self._open_file()
+            self._close_on_exit = True
+
+        # replace "x+" with "w+" since they are equivalent after the file is created
+        if self.mode == "x+":
+            self.mode = "w+"
+
+        # load file or write empty header if creating new file
+        self._initialize_structure()
 
     @mode.validator
     def check(self, attribute, value):
@@ -68,11 +82,12 @@ class SafeTensors:
     def readonly(self):
         return "+" not in self.mode
 
-    def __attrs_post_init__(self):
-        if not hasattr(self.file, "fileno"):
-            raise AssertionError
+    def __enter__(self):
+        return self
 
-        self._initialize_structure()
+    def __exit__(self, exc_type, exc_value, traceback):
+        if self._close_on_exit:
+            self.file.close()
 
     def _initialize_structure(self):
         if self.mode != "w+":
@@ -142,21 +157,19 @@ class SafeTensors:
         --------
 
         ::
-            st = SafeTensors(...)
-            st_numpy = st.as_numpy()
-
-            my_tensor = st_numpy["my_tensor_name"]
-            st_numpy.update({"new_tensor": my_array, "another_tensor": my_array_2})
-            del st_numpy["my_tensor_name"]
+            with SafeTensors(...) as st, st.as_numpy() as st_numpy:
+                do_stuff(st_numpy["my_tensor_name"] * 3 + 1)
+                st_numpy.update({"new_tensor": my_array, "another_tensor": my_array_2})
+                del st_numpy["my_tensor_name"]
         """
         from .numpy import NumpyAdapter
 
         return NumpyAdapter(self, **kw)
 
-    def as_pytorch(self, **kw):
-        raise NotImplementedError(
-            "pytorch support not implemented, why don't you do it and submit a patch?"
-        )
+    def as_buffer(self, **kw):
+        from .mmap import BufferAdapter
+
+        return BufferAdapter(self, **kw)
 
     def create_or_replace_or_delete_tensors(
         self, tensors: "Dict[str, Optional[SafeTensorInfo]]"
@@ -183,20 +196,18 @@ class SafeTensors:
         self.file.write(b"\0")
         self.file.flush()
         self._write_header(file_map)
+        self.truncate()
         self.sparsify()
 
     def truncate(self):
         """
         Truncate file to its smallest possible size.
         """
-        try:
-            last_tensor = self.tensors.values()[-1]
-        except IndexError:
-            eof = self.header_size
-        else:
-            eof = last_tensor.end
-
-        self.file.truncate(eof)
+        n = self.map.minimum_file_size
+        self.file.seek(n)
+        self.file.write(b"\0")
+        self.file.flush()
+        self.file.truncate(n)
 
     def sparsify(self):
         """
@@ -230,6 +241,7 @@ class SafeTensors:
 @attr.s
 class DType:
     name: str = attr.ib()
+    kind: str = attr.ib()
     size_in_bytes: int = attr.ib()
     numpy_name: str = attr.ib()
     recommended_alignment_in_bytes: int = attr.ib(default=None)
@@ -243,12 +255,12 @@ class DType:
 
 
 _dtypes_list = [
-    DType("BOOL", 1, "bool"),
-    DType("BF16", 2, "bfloat16"),
+    DType("BOOL", "b", 1, "bool"),
+    DType("BF16", "bf", 2, "bfloat16"),
 ]
-_dtypes_list += (DType(f"U{n}", n // 8, f"uint{n}") for n in (8, 16, 32, 64))
-_dtypes_list += (DType(f"I{n}", n // 8, f"int{n}") for n in (8, 16, 32, 64))
-_dtypes_list += (DType(f"F{n}", n // 8, f"float{n}") for n in (16, 32, 64))
+_dtypes_list += (DType(f"U{n * 8}", "u", n, f"uint{n * 8}") for n in (1, 2, 4, 8))
+_dtypes_list += (DType(f"I{n * 8}", "i", n, f"int{n * 8}") for n in (1, 2, 4, 8))
+_dtypes_list += (DType(f"F{n * 8}", "f", n, f"float{n * 8}") for n in (2, 4, 8))
 dtypes = {dt.name: dt for dt in _dtypes_list}
 del _dtypes_list
 
@@ -325,7 +337,7 @@ class SafeTensorsSchema(marshmallow.Schema):
     tensors = fields.Dict(
         fields.String(), fields.Nested(SafeTensorSchema), required=True
     )
-    metadata = fields.Raw(required=False)
+    metadata = fields.Dict(fields.String(), fields.String(), required=False)
 
     @marshmallow.pre_load
     def pre_load_metadata_split(self, data, **kwargs):
@@ -357,12 +369,13 @@ class SafeTensorsSchema(marshmallow.Schema):
         return merged
 
 
-@attr.s
-class FileMapSettings:
+@attr.s(eq=False)
+class Settings:
     header_size_alignment: int = attr.ib(default=4096)
     header_size_padding: int = attr.ib(default=1024)
     new_tensor_alignment: int = attr.ib(default=64)
     maximum_header_size: int = attr.ib(default=100_000_000)
+    maximum_file_size: int = attr.ib(2 ** 40)
 
 
 @attr.s
@@ -378,8 +391,8 @@ def sorted_tensors_dict(values=()):
 
 @attr.s
 class FileMap:
-    settings: FileMapSettings = attr.ib()
-    metadata: object = attr.ib(default=None)
+    settings: Settings = attr.ib()
+    metadata: Optional[Dict[str, str]] = attr.ib(default=None)
     tensors: "Dict[str, SafeTensorInfo]" = attr.ib(factory=sorted_tensors_dict)
     pending_moves: List[MoveRequest] = attr.ib(factory=list)
 
@@ -461,7 +474,7 @@ class FileMap:
         self._header = None
 
     @classmethod
-    def from_file(cls, f, settings: FileMapSettings, **kw):
+    def from_file(cls, f, settings: Settings, **kw):
         f.seek(0)
         length_data = f.read(8)
         if not length_data:
@@ -545,14 +558,19 @@ class FileMap:
             # The smallest possible header size would still exceed the available
             # space. Since tensors must be moved, make enough room so that
             # this movement doesn't need to happen again too soon.
-            header_size = align(
-                self.settings.header_size_alignment,
-                largest_json_header.unpadded_length + self.settings.header_size_padding,
+            header_size = self.compute_reserved_header_space(
+                largest_json_header.unpadded_length
             )
             final_header = self._get_header_for_header_size(header_size)
 
         return bytes(final_header)
 
+    def compute_reserved_header_space(self, header_size: int) -> int:
+        return align(
+            self.settings.header_size_alignment,
+            header_size + self.settings.header_size_padding,
+        )
+
     @property
     def header_size(self):
         return len(self.header)
@@ -563,18 +581,27 @@ class FileMap:
             start = tensor.start
             end = tensor.end
 
-            if start == end:
-                continue  # ignore empty tensors
-
-            if start < prev_end:
+            req_size = tensor.required_size
+            if req_size != end - start:
                 raise SafeTensorError(
-                    f"tensor {name!r} overlaps with previous structure"
+                    f"tensor {name!r} {tensor!r} size is wrong, expected {req_size}"
                 )
-            if end <= start:
-                raise SafeTensorError(f"tensor {name!r} does not obey `start < end`")
+
+            if start != end:
+                # only check tensors of nonzero size
+
+                if start < prev_end:
+                    raise SafeTensorError(
+                        f"tensor {name!r} overlaps with previous structure"
+                    )
 
             prev_end = end
 
+        if prev_end > self.settings.maximum_file_size:
+            raise SafeTensorError(
+                f"last tensor position {prev_end} exceeds maximum file size"
+            )
+
     @property
     def minimum_file_size(self):
         tensors = self.tensors
@@ -583,7 +610,7 @@ class FileMap:
 
     def get_empty_ranges(self):
         ranges = []
-        prev_end = self.header_size
+        prev_end = self.compute_reserved_header_space(self.header_size)
 
         for name, tensor in self.tensors.items():
             start = tensor.start
