@@ -1,9 +1,19 @@
+import os
+from typing import IO
+
+
 def align(alignment: int, position: int) -> int:
     assert alignment & (alignment - 1) == 0
     mask = alignment - 1
     return (position + mask) & ~mask
 
 
+def align_down(alignment: int, position: int) -> int:
+    assert alignment & (alignment - 1) == 0
+    mask = alignment - 1
+    return position & ~mask
+
+
 def samefile_copy(f, src: int, dst: int, length: int, block_size: int = 1 << 18):
     if src < dst:
         src += length
@@ -28,3 +38,6 @@ def samefile_copy(f, src: int, dst: int, length: int, block_size: int = 1 << 18)
             f.write(data)
             src += to_read
             dst += to_read
+
+def get_file_size(file: IO) -> int:
+    return os.fstat(file.fileno()).st_size
